from .loss_function import LossFunction
from .binary_crossentropy import BinaryCrossentropy
from .categorical_crossentropy import CategoricalCrossentropy
from .mse import MSE
from .mae import MAE
from .utils import get_loss_function



import numpy as np

from .loss_function import LossFunction


class MAE(LossFunction):
    """
    Class that calculates the Mean Absolute Error loss
    """
    def __init__(self):
        super(MAE, self).__init__(name="mae")

    @staticmethod
    def compute_loss(y_true: np.ndarray, y_pred: np.ndarray):
        """
        Calculates the Mean Absolute Error loss
        :param y_true: expected output
        :param y_pred: predictions
        :return: Mean Absolute Error loss
        """
        return np.abs(y_true - y_pred).mean()

    @staticmethod
    def gradient(y_true: np.ndarray, y_pred: np.ndarray):
        """
        Calculates the gradient of the Mean Absolute Error loss with respect to the predictions
        
        Args:
            y_true: labels
            y_pred: predictions
        
        """
        return np.sign(y_pred - y_true)



import numpy as np
from typing import Tuple

from .optimizer import Optimizer


class Adagrad(Optimizer):
    """
    Adagrad optimizer.
    """

    def __init__(self, lr: float = 0.01, eps: float = 1e-8):
        super(Adagrad, self).__init__(lr)
        self.eps = eps  # epsilon to avoid division by zero

    def update(self, parameters: Tuple[np.ndarray, np.ndarray], gradients: Tuple[np.ndarray, np.ndarray]):
        """

        Args:
            parameters: Parameters to update passed as a tuple of two numpy arrays composed of the weights and biases.
            gradients: Gradients of the parameters passed as a tuple of two numpy arrays composed of the delta of
                the weights and biases.

        """
        w, b = parameters
        dw, db = gradients

        







import numpy as np

from . import ActivationFunction


class ReLU(ActivationFunction):

    def __init__(self):
        super().__init__(name='relu')

    @staticmethod
    def forward(x: np.ndarray) -> np.ndarray:
        return x * (x > 0)

    @staticmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        return 1. * (z > 0)



import numpy as np

from .activation_function import ActivationFunction


class Softmax(ActivationFunction):
    """
    Softmax activation function.
    """

    def __init__(self):
        super().__init__(name="softmax")

    @staticmethod
    def forward(x: np.ndarray) -> np.ndarray:
        """
        Compute the softmax of the input.
        """
        exp_x = np.exp(x - np.max(x, axis=1, keepdims=True))
        return exp_x / np.sum(exp_x, axis=1, keepdims=True)

    @staticmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        """
        Compute the gradient of the softmax (d_a/d_z) of the input.
        Args:
            z: The input to the softmax function. It has shape (batch_size, num_classes).
        """
        # code adapted from:
        # https://stackoverflow.com/questions/36279904/softmax-derivative-in-numpy-approaches-0-implementation
        softmax = Softmax.forward(z)
        jacobian = - softmax[..., None] * softmax[:, None, :]  # off-diagonal Jacobian
        iy, ix = np.diag_indices_from(jacobian[0])
        jacobian[:, iy, ix] = softmax * (1. - softmax)  # diagonal

        # The code above is equivalent to the following, but is much faster (2x):
        # jacobian = np.empty((outputs.shape[0], outputs.shape[1], outputs.shape[1]))
        # softmax = Softmax.forward(outputs)
        # for m in range(outputs.shape[0]):
        #     for i in range(outputs.shape[1]):
        #         for j in range(outputs.shape[1]):
        #             jacobian[m, i, j] = softmax[m, i] * ((i == j) - softmax[m, j])

        return jacobian



import numpy as np
from typing import Union


from dlfs.convolutions import Convolutioner


class WinogradConvolutioner(Convolutioner):
    """A convolutioner that uses the Winograd convolution algorithm.

    The Winograd algorithm reduces the number of multiplication operations by transforming the input feature map and
    executing a series of transformations on the filter. The result is a much faster convolution.
    """

    def __init__(self,
                 image_size: Union[int, tuple],
                 kernel_size: Union[int, tuple],
                 padding: Union[int, tuple] = (0, 0),
                 stride: Union[int, tuple] = (1, 1)):

        super().__init__(image_size, kernel_size, padding, stride)

    @staticmethod
    def convolve_multichannel(image: np.ndarray,
                              kernel: np.ndarray,
                              padding: Union[int, tuple] = (0, 0),
                              stride: Union[int, tuple] = (1, 1),
                              using_batches: bool = False) -> np.ndarray:
        pass

    @staticmethod
    def convolve_grayscale(image: np.ndarray,
                           kernel: np.ndarray,
                           padding: Union[int, tuple] = (0, 0),
                           stride: Union[int, tuple] = (1, 1),
                           using_batches: bool = False) -> np.ndarray:
        pass



import numpy as np

from abc import ABC, abstractmethod


class ActivationFunction(ABC):
    """
    Base class for activation functions.
    """

    def __init__(self, name):
        self.name = name

    # Methods
    # -------------------------------------------------------------------------

    @staticmethod
    @abstractmethod
    def forward(x: np.ndarray) -> np.ndarray:
        """
        Forward pass of the activation function.

        Args:
            x: Input of the activation function.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        """
        Compute the gradient of the activation function.

        Args:
            z: Input of the activation function.
        """
        raise NotImplementedError

    def __call__(self, x):
        return self.forward(x)

    def __str__(self):
        return self.name



from . import MAE, MSE, BinaryCrossentropy, CategoricalCrossentropy


def get_loss_function(loss_name: str):
    if loss_name == 'mae':
        return MAE()
    elif loss_name == 'mse':
        return MSE()
    elif loss_name == 'binary_crossentropy':
        return BinaryCrossentropy()
    elif loss_name == 'categorical_crossentropy':
        return CategoricalCrossentropy()
    else:
        raise ValueError(f'Unknown loss function: {loss_name}')



# Source: On Improving the Numerical Stability of Winograd Convolutions
# By: Kevin Vincent and Kevin J. Stephano and Michael A. Frumkin and Boris Ginsburg and Julien Demouth
# (https://openreview.net/pdf?id=H1ZaRZVKg)

import numpy as np
from typing import Union, List, Tuple

np.set_printoptions(formatter={
    'float': lambda x: "{0:0.3f}".format(x)
})


def vandermonde_matrix(b: int, points: List[Tuple[Union[int, float]]]) -> np.ndarray:
    """
    This is equivalent to V_{axb} on the document.
    This function returns a trimmed Vandermonde Matrix of size len(points) x b

        f_0^0 * g_0^b-1       f_0^1 * g_0^b-2       ... f_0^b-1 * g_0^0
        .
        .
        f_a-1^0 * g_a-1^b-1   f_a-1^1 * g_a-1^b-2   ... f_a-1^b-1 * g_a-1^0

    Arguments:
        b (int) : One of the dimensions of the matrix
        points (list[tuple[Union[int, float]]]) : List of homogeneous coordinates
    Returns:
        V (np.ndarray[Union[int, float]]) : The trimmed Vandermonde matrix
    """
    return np.array([[i[0] ** j * i[1] ** (b - j - 1) for j in range(b)] for i in points])


def gen_points(num_points: int) -> List[Tuple[float, int]]:
    """
    This function generates homogeneous coordinates, starting from (0, 1), then, (1/2, 1), then (-1/2, 1), and the infinite point (0, 1).

        [(f_0, g_0), ..., (f_a-1, g_a-1)]

    Args:
        num_points (int): The desired number of points

    Returns:
        list[tuple[Union[int, float]]]: List of homogeneous coordinates
    """
    points = [(0, 1)]
    if num_points % 2:
        points.extend([(y / 2, 1) for x in range(1, num_points // 2) for y in (x, -x)])
        points.extend([(num_points // 2 / 2, 1), (1, 0)])
    else:
        points.extend([(y / 2, 1) for x in range(1, num_points // 2) for y in (x, -x)])
        points.append((1, 0))
    return points


def winograd_get_matrices(m: int, n: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    This is equivalent to F(m, n) on the paper.
    This function returns the matrices necessary for calculating the Winograd convolution given the dimension of the filter and the dimension of the output. With those, we can infer the size of the input.
    The usage of these matrices depends on the dimension of the convolution. If we denote x as the signal and w as the filter:
    - If it is 1D: F(m, n) = x * w = Y.T @ [(X.T @ x) * (W @ w)]
    - If it is 2D: F(m x i, n x j) = x * w = Y.T @ [(X.T @ x @ X') * (W @ w @ W'.T)] @ Y'
    In the latter case, if m = i and n = j, then Y = Y', X = X', W = W'. If not, then you would have to calculate both F(m, n) to get Y, X, W and F(i, j) to get Y', X', W'.

    Args:
        m (int): The size of the output
        n (int): The size of the filter

    Returns:
        A tuple with the aforementioned Y, X, W in this order.
    """
    num_points = m + n - 1
    points = gen_points(num_points)
    Y = vandermonde_matrix(m, points)
    X = np.linalg.inv(vandermonde_matrix(num_points, points))
    W = vandermonde_matrix(n, points)
    return Y, X, W


if __name__ == "__main__":
    Y, X, W = winograd_get_matrices(3, 3)
    print('Y:\n', Y)
    print('X:\n', X)
    print('W:\n', W)



from typing import Union

from . import ActivationFunction, Sigmoid, ReLU, Softmax, Linear, Tanh


def get_activation_function(activation_function_name: Union[str, None]) -> Union[ActivationFunction, None]:
    """
    Returns an activation function object based on the name of the activation function.
    Args:
        activation_function_name: The name of the activation function. The possible names are:
            - "relu": ReLU
            - "softmax": Softmax
            - "sigmoid": Sigmoid
            - "tanh": Tanh
            - "linear": Linear
            - None: No activation function (i.e. the input is passed through, the same as Linear)
    Returns:
        An activation function object or None if there is no activation function (the same as using 'Linear').
    """
    if activation_function_name is None:
        return None

    if activation_function_name == 'relu':
        return ReLU()
    elif activation_function_name == 'softmax':
        return Softmax()
    elif activation_function_name == 'sigmoid':
        return Sigmoid()
    elif activation_function_name == 'linear':
        return Linear()
    elif activation_function_name == 'tanh':
        return Tanh()
    else:
        raise ValueError(f'Unknown activation function name: {activation_function_name}')



import numpy as np

from .loss_function import LossFunction


class MSE(LossFunction):
    """
    Class that defines the Mean Squared Error loss function
    """

    def __init__(self):
        super(MSE, self).__init__(name="mse")

    @staticmethod
    def compute_loss(y_true, y_pred):
        """
        Calculates the Mean Squared Error loss
        """
        return np.sum(np.square(y_true - y_pred)) / (2 * y_true.shape[0])

    @staticmethod
    def gradient(y_true, y_pred):
        """
        Calculates the gradient of the Mean Squared Error loss
        """
        return y_pred - y_true



import numpy as np
from typing import Tuple

from dlfs.optimizers import Optimizer
from dlfs.layers import Layer


class SGD(Optimizer):
    """
    Mini-batch stochastic gradient descent.
    """

    def __init__(self, learning_rate: float = 0.01):
        super(SGD, self).__init__(learning_rate=learning_rate)

    def update(self, layer: Layer, gradients: Tuple[np.ndarray, np.ndarray]):
        """
        Update parameters using SGD

        Args:
            layer: The layer to update.
            gradients: delta of parameters (weights and bias)
        """
        dw, db = gradients

        layer.__weights -= self.learning_rate * dw
        layer.__bias -= self.learning_rate * db

    def add_slot(self, layer: Layer):
        """
        Add a slot for the layer.
        """
        pass



from abc import ABC, abstractmethod
import numpy as np


class Metric(ABC):

    def __init__(self, name):
        self.name = name

    @staticmethod
    @abstractmethod
    def compute_metric(y_true: np.ndarray, y_pred: np.ndarray) -> float:
        pass

    def __str__(self):
        return self.name

    def __call__(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        return self.compute_metric(y_true, y_pred)



import numpy as np


def train_test_split(x: np.ndarray, y: np.ndarray, test_size: float = 0.3, random_state: int = None):
	"""
	Split the dataset into train and test sets.
	Args:
		x: input data
		y: labels
		test_size: the proportion of the dataset to include in the test split
		random_state: the seed used by the random number generator

	Returns:
		x_train: training data
		y_train: training labels
		x_test: testing data
		y_test: testing labels
	"""
	train = x.shape[0] - int(x.shape[0] * test_size)
	arr = np.arange(x.shape[0])
	np.random.RandomState(seed=random_state).shuffle(arr)
	return x[arr[:train]], x[arr[train:]], y[arr[:train]], y[arr[train:]]



from . import Optimizer, SGD, SGDMomentum, Adam, Adagrad


def get_optimizer(optimizer_name: str) -> Optimizer:
    """
    Get an optimizer by name.

    Args:
        optimizer_name (str): The name of the optimizer. Valid names are:
            - sgd
            - sgd_momentum
            - adam
            - adagrad

    Returns:
        The optimizer.
    """
    if optimizer_name == "sgd":
        return SGD()
    elif optimizer_name == "sgd_momentum":
        return SGDMomentum()
    elif optimizer_name == "adam":
        return Adam()
    elif optimizer_name == "adagrad":
        return Adagrad()
    else:
        raise ValueError(f"Unknown optimizer name: {optimizer_name}")



import numpy as np

from .activation_function import ActivationFunction


class Sigmoid(ActivationFunction):

    def __init__(self):
        super().__init__(name='sigmoid')

    @staticmethod
    def forward(x: np.ndarray) -> np.ndarray:
        return 1 / (1 + np.exp(-x))

    @staticmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        sigmoid = Sigmoid.forward(z)
        return sigmoid * (1 - sigmoid)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the Sequential model class."""


from collections import deque
from copy import deepcopy
import numpy as np
from typing import List, Dict, Tuple
from tqdm import tqdm
import pickle


from . import Model
from dlfs.layers import Layer
from dlfs.optimizers import Optimizer, get_optimizer
from dlfs.losses import LossFunction, get_loss_function
from dlfs.metrics import Metric, get_metric
from dlfs.preprocessing import train_test_split


class Sequential(Model):
    """A sequential model is a linear stack of layers.

    Args:
        layers (List[Layer]): A list of layers.
        name (str): The name of the model.

    Attributes:
        layers (List[Layer]): A list of layers.
        name (str): The name of the model.
        optimizer (Optimizer): The optimizer used to optimize the model.
        loss (LossFunction): The loss function used to compute the loss.
        metrics (Dict[str, Union[Metric, LossFunction]]): A dictionary of metrics.
        __trainable (bool): Whether the model is trainable.

    Usage:
        >>> from dlfs.layers import Dense

        >>> model = Sequential([Dense(10, activation='relu'), Dense(1)])

        >>> model.summary()

        >>> model.compile(optimizer='sgd', loss='mse', metrics=['mae'])

        >>> model.fit(x_train, y_train, epochs=10)


    """

    layers: List[Layer]
    name: str
    loss: LossFunction or None
    optimizer: Optimizer or None
    metrics: Dict[str, Metric] or None
    trainable: bool

    def __init__(self, layers: List[Layer] = None, name: str = "Sequential"):

        super().__init__(name=name)

        # check that the first layers has set the input shape
        if layers and layers[0].input_shape is None:
            raise ValueError("The first layer must have an input shape")
        self.layers = []
        if layers is not None:
            for layer in layers:
                self.add(layer)
        self.name = name
        self.loss = None
        self.optimizer = None
        self.metrics = None
        self.__trainable = True
        self.__counter = 1

    @property
    def trainable(self):
        return self.__trainable

    @trainable.setter
    def trainable(self, value: bool):
        self.__trainable = value
        for layer in self.layers:
            layer.trainable = value

    def add(self, layer: Layer):
        """Adds a layer to the model

        Args:
            layer (Layer): the layer to add
        """

        # check that the first layers has set the input shape
        if not self.layers and layer.input_shape is None:
            raise ValueError("The first layer must have an input shape")

        # initialize the layer
        if not layer.initialized:
            if layer.input_shape is None:
                layer.initialize(input_shape=self.layers[-1].output_shape)
            else:
                layer.initialize(input_shape=layer.input_shape)

        # each layer must have an unique name
        layer.name = f"{layer.name}{self.__counter}"
        self.__counter += 1

        self.layers.append(layer)

    def concatenate(self, model: 'Sequential'):
        """Concatenates two models.
        
        Args:
            model (Sequential): the model to concatenate
        """
        if self.layers[-1].output_shape != model.layers[0].input_shape:
            raise ValueError("The two models must have the same input shape")
        self.layers += model.layers

    def compile(self, optimizer: Optimizer or str, loss: LossFunction or str, metrics: List[str] = None):
        """
        Compile the model

        Args:
            optimizer: the optimizer to use
            loss: the loss function to use
            metrics: the metrics to use. The metrics by default are the ones used by the loss function
            Allowed values are:
                - "accuracy"
        """
        self.optimizer = get_optimizer(optimizer) if isinstance(optimizer, str) else optimizer
        self.loss = get_loss_function(loss) if isinstance(loss, str) else loss
        self.metrics = {} if metrics is None else {metric: get_metric(metric) for metric in metrics}

        # check that each layer has an unique name
        names = set()
        for layer in self.layers:
            if layer.name in names:
                raise ValueError(f"The layer name {layer.name} is not unique")
            names.add(layer.name)
            self.optimizer.add_slot(layer)

    def summary(self):
        """
        Print the summary of the model as a table
        """
        print(f"Model: {self.name}")
        print("-" * len(self.name))
        print("Layers:")
        for layer in self.layers:
            print("-" * len(layer.summary()))
            print(f"{layer.summary()}")

        print("-" * len(self.name))
        print("Total params:", sum(layer.count_params() for layer in self.layers))
        print("Trainable params:", sum(layer.count_params() for layer in self.layers if layer.trainable))
        print("-" * len(self.name))

    def get_deltas(self, y_pred: np.ndarray, y_true: np.ndarray) -> deque:
        """
        Backpropagation of the loss function
        Args:
            y_pred: the predictions
            y_true: the true labels

        Returns:
            the deltas of the loss function (as a deque)
        """

        # initialize the deltas
        deltas = deque()

        last_layer = self.layers[-1]
        deltas.appendleft(last_layer.get_delta(self.loss.gradient(y_true, y_pred)))

        # backward pass
        for layer in reversed(self.layers[:-1]):
            d_inputs = last_layer.get_d_inputs(deltas[0])
            delta = layer.get_delta(d_inputs)
            deltas.appendleft(delta)
            last_layer = layer

        return deltas

    def train_on_batch(self, x_batch: np.ndarray, y_batch: np.ndarray) -> Dict[str, float]:
        """
        Train the model on a batch of data

        Args:
            x_batch: the batch of inputs
            y_batch: the batch of expected outputs

        Returns:
            the loss and the metrics
        """

        # forward pass
        y_pred = self.predict(x_batch, training=True)

        # backward pass: get the deltas
        deltas: deque = self.get_deltas(y_pred, y_batch)

        # backward pass: update the weights
        for i, layer in enumerate(self.layers):
            if layer.trainable:
                layer.update(self.optimizer, deltas[i])

        # compute the loss for the batch
        loss = self.loss.compute_loss(y_batch, y_pred)

        # compute the metrics for the batch
        metrics = {metric.name: metric(y_batch, y_pred) for metric in self.metrics.values()}

        return {**{'loss': loss}, **metrics}

    def __check_data(self, x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:

        if x.shape[1:] != self.layers[0].input_shape[1:]:
            raise ValueError(f"The input shape of the model is {self.layers[0].input_shape}")

        if y.ndim == 1:
            y = y.reshape(-1, 1)

        # check if the labels are valid
        if y.shape[1:] != self.layers[-1].output_shape[1:]:
            raise ValueError(f"The output shape of the model is {self.layers[-1].output_shape}")

        return x, y

    def __update_io_shapes(self, batch_size: int):
        """
        Update the input and output shapes of the layers to match the batch size
        Args:
            batch_size: the batch size
        """
        for layer in self.layers:
            layer.input_shape = (batch_size, *layer.input_shape[1:])
            layer.output_shape = (batch_size, *layer.output_shape[1:])

    def __print_results(self,
                        metrics: Dict[str, float],
                        val_metrics: Dict[str, float] = None,
                        title: str = "Epoch",
                        progress: int = 0,
                        total: int = 0,
                        history: Dict[str, list] = None) -> None:
        """
        Print the results of the training and update the history
        Args:
            metrics: the metrics (and loss) of the current epoch
            val_metrics: the metrics of the validation set
            title: the title of the printed results (Batch or Epoch)
            progress: the progress of the training. The number of batches processed or the number of epochs
            total: the total number of batches or epochs
            history: the history of the training to update
        """

        using_validation = val_metrics is not None
        if using_validation:
            # print the results
            print(f"{title} ({progress}/{total}) - loss: {metrics['loss']} - val_loss: {val_metrics['val_loss']}")
            for metric in self.metrics:
                print(f"- {metric}: {metrics[metric]} - val_{metric}: {val_metrics['val_' + metric]}")
        else:
            # print the results
            print(f"{title} ({progress}/{total}) - loss: {metrics['loss']}")
            for metric in self.metrics:
                print(f"- {metric}: {metrics[metric]}")

        if history is not None:
            for metric in metrics:
                history[metric].append(metrics[metric])
            if using_validation:
                for metric in val_metrics:
                    history[metric].append(val_metrics[metric])
        print("")

    def fit(self,
            x: np.ndarray,
            y: np.ndarray,
            epochs: int = 1,
            batch_size: int = 32,
            verbose: int = 1,
            validation_data: Tuple[np.ndarray, np.ndarray] = None,
            validation_split: float = 0.,
            shuffle: bool = True,
            initial_epoch: int = 0) -> Dict[str, List[float]]:
        """
        Fit the model to the data

        Args:
            x: the input data. With shape (n_samples, n_features)
            y: the labels
            epochs: the number of epochs to train the model
            batch_size: the batch size
            verbose: the verbosity mode:
                * 0 doesn't show anything.
                * 1 prints one line per batch.
                * 2 prints one line per epoch.
                * 3 shows the progress bar (per epoch).
            validation_data: the validation data
            validation_split: the validation split
            shuffle: whether to shuffle the data
            initial_epoch: the initial epoch
        """
        # CHECKS:
        # check if the model has been compiled
        if self.optimizer is None or self.loss is None:
            raise ValueError("You must compile the model before training")

        # check if the data dimensions are valid
        x, y = self.__check_data(x, y)

        using_validation_data = True
        # if validation data is not provided, split the data into train and validation
        if validation_split > 0:
            if validation_data is not None:
                raise ValueError("validation_data and validation_split cannot be used together")
            x, x_val, y, y_val = train_test_split(x, y, test_size=validation_split)
        elif validation_data is not None:
            x_val, y_val = validation_data

            # check if the data dimensions are valid
            x_val, y_val = self.__check_data(x_val, y_val)
        else:
            x_val = None
            y_val = None
            using_validation_data = False

        if using_validation_data and verbose == 0:
            raise ValueError("validation_data and verbose=0 cannot be used together")

        # INITIALIZATION:
        # --------------------------------------------------
        np.set_printoptions(precision=4)

        # Update the input_shape of the layers to take into account the batch_size
        for layer in self.layers:
            layer.input_shape = (batch_size, *layer.input_shape[1:])

        # Update the output_shape of the layers to take into account the batch_size
        for layer in self.layers:
            layer.output_shape = (batch_size, *layer.output_shape[1:])

        # initialize the history
        history = {"loss": [], "val_loss": []}
        for metric in self.metrics:
            history[metric] = []
            history["val_" + metric] = []

        # set the optimizers
        for layer in self.layers:
            layer.optimizer = deepcopy(self.optimizer)

        # TRAINING:
        # --------------------------------------------------
        # create the progress bar if verbose is 0
        r = range(initial_epoch, epochs)
        range_epochs = r if verbose != 3 else tqdm(r)

        # loop over the number of epochs
        for epoch in range_epochs:
            # initialize the total loss for the epoch if verbose is 2

            epoch_metrics = {metric: 0.0 for metric in self.metrics}
            epoch_metrics['loss'] = 0.0

            val_metrics = None

            # loop over the data in batches
            total_data_used_per_epoch = 0
            generator = self.batch_generator(x, y, batch_size, shuffle)
            if verbose == 2:
                generator = tqdm(generator, total=len(x) // batch_size)

            for x_batch, y_batch in generator:

                total_data_used_per_epoch += x_batch.shape[0]

                metrics = self.train_on_batch(x_batch, y_batch)

                # update the total loss and metrics for the epoch
                for metric in epoch_metrics:
                    epoch_metrics[metric] += metrics[metric]

                if using_validation_data and verbose == 1:
                    # compute the loss and the metrics for the validation data
                    y_val_pred = self.predict(x_val)
                    val_loss = self.loss.compute_loss(y_val, y_val_pred)
                    val_metrics = {"val_" + metric: self.metrics[metric](y_val, y_val_pred)
                                   for metric in self.metrics}

                    # add val_loss to val_metrics
                    val_metrics["val_loss"] = val_loss

                # print the loss and the metrics per batch (if verbose is 1)
                if verbose == 1:
                    self.__print_results(metrics, val_metrics, "Batch",
                                         progress=total_data_used_per_epoch, total=len(x))

            # print the metrics of the epoch (if verbose is 1 or 2)
            if verbose == 1 or verbose == 2:

                # compute the average loss and metrics for the epoch
                epoch_metrics = {metric: epoch_metrics[metric] / (len(x) // batch_size) for metric in epoch_metrics}
                val_epoch_metrics = None
                if using_validation_data:
                    # compute the loss and the metrics for the validation data
                    y_val_pred = self.predict(x_val)
                    val_epoch_loss = self.loss.compute_loss(y_val, y_val_pred)
                    val_epoch_metrics = {"val_" + metric: self.metrics[metric](y_val, y_val_pred)
                                         for metric in self.metrics}

                    # add val_loss to val_metrics
                    val_epoch_metrics["val_loss"] = val_epoch_loss

                # print the metrics of the epoch
                self.__print_results(epoch_metrics, val_epoch_metrics, "Epoch",
                                     progress=epoch + 1, total=epochs, history=history)

            # save the history
            for metric in epoch_metrics:
                history[metric].append(epoch_metrics[metric] / (len(x) // batch_size))

        if verbose == 3:
            # print final results
            self.evaluate(x, y)
            if using_validation_data:
                self.evaluate(x_val, y_val, prefix="val_")
        return history

    @staticmethod
    def batch_generator(x: np.ndarray, y: np.ndarray, batch_size: int, shuffle: bool = True):
        """
        Generates batches of data
        Args:
            x: the input data
            y: the labels
            batch_size: the batch size
            shuffle: whether to shuffle the data

        Yields:
            x_batch: the batch of input data
            y_batch: the batch of labels
        """
        # get the number of batches
        n_batches = len(x) // batch_size

        if shuffle:
            # shuffle the data
            indices = np.arange(len(x))
            np.random.shuffle(indices)
            x = x[indices]
            y = y[indices]

        # loop over the batches
        for i in range(0, n_batches * batch_size, batch_size):
            # get the batch data
            x_batch = x[i:i + batch_size]
            y_batch = y[i:i + batch_size]
            # yield the batch
            yield x_batch, y_batch

    def evaluate(self,
                 x: np.ndarray,
                 y: np.ndarray,
                 batch_size: int = None,
                 verbose: int = 1,
                 prefix: str = "") -> dict:
        """
        Args:
            x: the input data
            y: the labels
            batch_size: the batch size (default: len(x))
            verbose: the verbosity mode (0, 1, or 2)
            prefix: the prefix to print

        Returns:
            The average loss and the average metrics of the model on the given data
        """
        # check the data
        x, y = self.__check_data(x, y)

        # get the batch size
        batch_size = batch_size or len(x)

        # update input and output shapes of the model
        self.__update_io_shapes(batch_size)

        # initialize the metrics
        avg_loss = 0.0
        avg_metrics = {prefix + metric: 0.0 for metric in self.metrics}

        n_batches = len(x) // batch_size

        # loop over the batches
        for i in range(0, n_batches * batch_size, batch_size):
            # get the batch data
            y_pred = x[i:i + batch_size]
            y_batch = y[i:i + batch_size]

            # forward pass
            y_pred = self.predict(y_pred)

            # compute the loss for the batch
            loss = self.loss.compute_loss(y_batch, y_pred)

            # compute the metrics for the batch
            metrics = {prefix + metric: self.metrics[metric](y_batch, y_pred) for metric in self.metrics}

            # update the loss and the metrics
            avg_loss += 1/(i + 1) * (loss - avg_loss)
            for metric in self.metrics:
                avg_metrics[prefix + metric] += 1/(i + 1) * (metrics[prefix + metric] - avg_metrics[prefix + metric])

            # print the loss and the metrics per batch (if verbose is 1)
            if verbose == 2 and batch_size != len(x):
                print(f"Batch ({i // batch_size + 1}/{n_batches}) - loss: {loss}")
                print(f"\t{', '.join([prefix + f'{metric}: {metrics[prefix + metric]:.4f}' for metric in self.metrics])}")

        # print final loss and metrics
        if verbose > 0:
            print(prefix + f"loss: {avg_loss}")
            print(f"{', '.join([prefix + f'{metric}: {avg_metrics[prefix + metric]:.4f}' for metric in self.metrics])}")

        return {prefix + 'loss': avg_loss, **avg_metrics}

    def predict(self, x: np.ndarray, training: bool = False) -> np.ndarray:
        """

        Args:
            x : the input data
            training: whether the model is in training mode
        Returns:
            y_pred : the predictions
        """

        if not training:
            self.__update_io_shapes(len(x))

        last_input = x
        for layer in self.layers:
            last_input = layer.forward(last_input, training)
        return last_input

    def set_weights(self, weights: list, layers: list = None, only_weights: bool = False, only_biases: bool = False):
        """
        Sets the weights and biases of the model. The weights and biases are in the same order as the layers.
        Args:
            weights: the weights of the model (list of numpy arrays) [layer_0_weights, layer_0_bias, ...]
                The weights and biases of all the layers that have them must be included in the list
            layers: the layers of the model (list of Layer objects) which you want to set the weights of
                (default: all layers with weights)
            only_weights: whether to set only the weights. If True, the weights list must contain
                only the weights of the layers in the layers list. (default: False)
            only_biases: whether to set only the biases. If True, the weights list must contain
                only the biases of the layers in the layers list. (default: False)
        """
        layers_with_weights = layers or [layer for layer in self.layers if layer.has_weights]

        if only_weights:
            for i, w in enumerate(weights):
                layers_with_weights[i].set_weights(w)
        elif only_biases:
            for i, b in enumerate(weights):
                layers_with_weights[i].set_weights(bias=b)
        else:
            for i, layer in enumerate(layers_with_weights):
                layer.set_weights(weights[i], weights[i + len(layers_with_weights)])

    def save(self, path: str):
        """
        Saves the model to the given path using the pickle module
        Args:
            path: the path to save the model
        """
        with open(path, 'wb') as f:
            pickle.dump(self, f)

    @staticmethod
    def load(path: str):
        """
        Loads the model from the given path using the pickle module
        Args:
            path: the path to load the model
        """
        with open(path, 'rb') as f:
            return pickle.load(f)

    def __str__(self):
        return f"{self.__class__.__name__}({', '.join([str(layer) for layer in self.layers])})"

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return len(self.layers)

    def __getitem__(self, index: int):
        return self.layers[index]

    def __iter__(self):
        return iter(self.layers)

    def __add__(self, other: 'Sequential'):
        return Sequential(self.layers + other.layers)

    def __iadd__(self, other: 'Sequential'):
        self.concatenate(other)

    def __copy__(self):
        return deepcopy(Sequential(self.layers))


from abc import ABC, abstractmethod
from numpy import ndarray


class LossFunction(ABC):
    """
    Base class for loss functions.
    """

    def __init__(self, name):
        self.name = name

    @staticmethod
    @abstractmethod
    def compute_loss(y_true, y_pred) -> float:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def gradient(y_true, y_pred) -> ndarray:
        raise NotImplementedError

    def __call__(self, y_true, y_pred):
        return self.compute_loss(y_true, y_pred)



import numpy as np

from .loss_function import LossFunction


class BinaryCrossentropy(LossFunction):

    def __init__(self, name='binary_crossentropy'):
        super().__init__(name)

    @staticmethod
    def compute_loss(y_true, y_pred):
        """
        Compute the binary cross entropy loss.
        """
        y_pred_clipped = np.clip(y_pred, 1e-7, 1 - 1e-7)
        return np.mean(-y_true * np.log(y_pred_clipped) - (1 - y_true) * np.log(1 - y_pred_clipped))

    @staticmethod
    def gradient(y_true, y_pred):
        """
        Compute the gradient of the binary cross entropy loss.
        """
        return y_pred - y_true



import numpy as np
from typing import Union

from dlfs.convolutions import Convolutioner


class SimpleConvolutioner(Convolutioner):
    """
    A simple convolutioner that performs a convolution on a single image.

    Usage:
        >>> img = np.array([[2, 2, 1, 3],
                            [0, 3, 2, 1],
                            [1, 1, 0, 2],
                            [2, 0, 0, 1]])

        >>> k = np.array([[1,  0],
                          [2, -2]])
        >>> conv = SimpleConvolutioner(img.shape, k.shape)
        >>> print(conv.convolve(img, k, using_batches=False))
    """

    def __init__(self,
                 image_size: Union[int, tuple],
                 kernel_size: Union[int, tuple],
                 padding: Union[int, tuple] = (0, 0),
                 stride: Union[int, tuple] = (1, 1)):

        super().__init__(image_size, kernel_size, padding, stride)

    @staticmethod
    def convolve_grayscale(image: np.ndarray,
                           kernel: np.ndarray,
                           padding: Union[int, tuple] = (0, 0),
                           stride: int = 1,
                           using_batches: bool = False) -> np.ndarray:
        """
        Performs a valid convolution on an image (with only a channel) with a kernel.

        Args:
            image: A grayscale image.
            kernel: A kernel.
            padding: Whether to pad the image.
            stride: convolution stride size.
            using_batches: Whether to use batches.

        Returns:
            A grayscale image.
        """

        # Pad the image if padding is enabled
        padding = (padding, padding) if isinstance(padding, int) else padding
        image = Convolutioner.pad_image(image, padding) if padding != (0, 0) else image
        if using_batches:
            return np.array([SimpleConvolutioner.convolve_grayscale(image_batch, kernel, 0, stride)
                             for image_batch in image])

        # Get the dimensions of the image and kernel
        image_height, image_width = image.shape
        kernel_height, kernel_width = kernel.shape

        # Create the output image
        stride = (stride, stride) if isinstance(stride, int) else stride
        output_height = (image_height - kernel_height) // stride[0] + 1
        output_width = (image_width - kernel_width) // stride[1] + 1
        convolved_image = np.zeros((output_height, output_width))

        # Perform the convolution
        for i in range(output_height):
            for j in range(output_width):
                convolved_image[i, j] = np.sum(image[i * stride[0]:i * stride[0] + kernel_height,
                                                     j * stride[1]:j * stride[1] + kernel_width] * kernel)
        return convolved_image

    @staticmethod
    def convolve_multichannel(image: np.ndarray,
                              kernel: np.ndarray,
                              padding: Union[int, tuple] = (0, 0),
                              stride: Union[int, tuple] = (1, 1),
                              using_batches: bool = False) -> np.ndarray:
        """
        Performs a valid convolution on an image (with multiple channels) with a kernel.

        Args:
            image: A multichannel image.
            kernel: A kernel.
            padding: Whether to pad the image.
            stride: convolution stride size.
            using_batches: Whether to use batches.

        Returns:
            A multichannel image.
        """
        # Pad the image if padding is enabled
        padding = (padding, padding) if isinstance(padding, int) else padding
        image = Convolutioner.pad_image(image, padding) if padding != (0, 0) else image

        if using_batches:
            return np.array([SimpleConvolutioner.convolve_multichannel(image_batch, kernel, padding=0, stride=stride)
                             for image_batch in image])

        # Get the dimensions of the image and kernel
        image_height, image_width, image_channels = image.shape
        kernel_height, kernel_width, kernel_channels = kernel.shape

        # Create the output image
        stride = (stride, stride) if isinstance(stride, int) else stride
        output_height = (image_height - kernel_height) // stride[0] + 1
        output_width = (image_width - kernel_width) // stride[1] + 1
        convolved_image = np.zeros((output_height, output_width, kernel_channels))

        # Perform the convolution
        for i in range(output_height):
            for j in range(output_width):
                for k in range(kernel_channels):
                    convolved_image[i, j, k] = np.sum(image[i * stride[0]:i * stride[0] + kernel_height,
                                                            j * stride[1]:j * stride[1] + kernel_width,
                                                            k] * kernel[:, :, k])

        return convolved_image



import numpy as np

from . import ActivationFunction


class Linear(ActivationFunction):

    def __init__(self):
        super().__init__(name='linear')

    @staticmethod
    def forward(x):
        return x

    @staticmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        return np.ones_like(z)



import numpy as np

from .optimizer import Optimizer


class Adam(Optimizer):
    """
    Adam optimizer.
    """

    def __init__(self, learning_rate=0.001, decay=0., beta1=0.9, beta2=0.999, epsilon=1e-7):

        super(Adam, self).__init__(learning_rate)
        self.__current_learning_rate = learning_rate
        self.__beta1 = beta1
        self.__beta2 = beta2
        self.__epsilon = epsilon
        self.__iterations = 0
        self.__decay = 0
        self.init_layer()

    def init_layer(self):
        self.__layers = []
        for layer in sequential.layers:
            ...

    def update(self, layer):
        if not hasattr(layer, 'weight_cache'):
            layer.weight_momentums = np.zeros_like(layer.__weights)
            layer.weight_cache = np.zeros_like(layer.__weights)
            layer.bias_momentums = np.zeros_like(layer.__bias)
            layer.bias_cache = np.zeros_like(layer.__bias)

        layer.weight_momentums = self.__beta1 * layer.weight_momentums + (1 - self.__beta1) * layer.dweights

    # Call once before any parameter updates




from typing import Union

from . import Convolutioner, SimpleConvolutioner, WinogradConvolutioner


def get_convolution(convolution_type: str,
                    image_size: Union[int, tuple],
                    kernel_size: Union[int, tuple],
                    padding: tuple = (0, 0),
                    stride: Union[int, tuple] = (1, 1)) -> Convolutioner:
    """
    Returns an activation function object based on the name of the activation function.
    Args:
        convolution_type: The name of the activation function. The possible names are:
            - "simple": Simple convolution.
            - "winograd": Winograd convolution.
        image_size: The size of the matrix.
        kernel_size: The size of the kernel.
        padding: Whether to use padding or not.
        stride: The stride of the convolution.

    Returns:
        An activation function object or None if there is no activation function (the same as using 'Linear').
    """

    if convolution_type == "simple":
        return SimpleConvolutioner(image_size, kernel_size, padding, stride)
    elif convolution_type == "winograd":
        return WinogradConvolutioner(image_size, kernel_size, padding, stride)
    else:
        raise ValueError(f'Unknown activation function name: {convolution_type}')


from typing import Union

from dlfs.metrics import Metric, Accuracy
from dlfs.losses import *


def get_metric(name: str) -> Union[Metric, LossFunction]:
    """
    Returns a metric object by name.
    """

    if name == 'accuracy':
        return Accuracy()
    elif name == 'mae':
        return MAE()
    elif name == 'mse':
        return MSE()
    else:
        raise ValueError(f'Unknown metric: {name}')



import numpy as np

from .activation_function import ActivationFunction


class Tanh(ActivationFunction):

    def __init__(self):
        super().__init__(name='tanh')

    @staticmethod
    def forward(x: np.ndarray) -> np.ndarray:
        return np.tanh(x)

    @staticmethod
    def gradient(z: np.ndarray) -> np.ndarray:
        return 1 - np.square(z)



from abc import ABC, abstractmethod
import numpy as np
from typing import Union


class Convolutioner(ABC):
    """
    Abstract class for convolutioners.
    """

    def __init__(self,
                 image_size: Union[int, tuple],
                 kernel_size: Union[int, tuple],
                 padding: Union[int, tuple] = (0, 0),
                 stride: Union[int, tuple] = (1, 1)):

        self.image_size = image_size if isinstance(image_size, tuple) else (image_size, image_size)
        self.kernel_size = kernel_size if isinstance(kernel_size, tuple) else (kernel_size, kernel_size)
        self.padding = padding
        self.stride = (stride, stride) if isinstance(stride, int) else stride

    @staticmethod
    @abstractmethod
    def convolve_grayscale(image: np.ndarray,
                           kernel: np.ndarray,
                           padding: Union[int, tuple] = (0, 0),
                           stride: Union[int, tuple] = (1, 1),
                           using_batches: bool = False) -> np.ndarray:
        pass

    @staticmethod
    @abstractmethod
    def convolve_multichannel(image: np.ndarray,
                              kernel: np.ndarray,
                              padding: tuple = (0, 0),
                              stride: Union[int, tuple] = (1, 1),
                              using_batches: bool = False) -> np.ndarray:
        pass

    def convolve(self,
                 image: np.ndarray,
                 kernel: np.ndarray,
                 using_batches: bool = True) -> np.ndarray:

        if using_batches:
            if image.ndim == 4:
                return self.convolve_multichannel(image, kernel, self.padding, self.stride, using_batches)
            elif image.ndim == 3:
                return self.convolve_grayscale(image, kernel, self.padding, self.stride, using_batches)
        else:
            if image.ndim == 2:
                return self.convolve_grayscale(image, kernel, self.padding, self.stride, using_batches=False)
            elif image.ndim == 3:
                return self.convolve_multichannel(image, kernel, self.padding, self.stride, using_batches=False)

        raise ValueError("Image must be 2D or 3D.")

    @staticmethod
    def pad_image(image: np.ndarray, padding: Union[int, tuple], using_batches: bool = False) -> np.ndarray:

        padding = (padding, padding) if isinstance(padding, int) else padding

        if not using_batches:
            if image.ndim == 3:
                image = np.pad(image, ((0, 0), (padding[0], padding[0]), (padding[1], padding[1])),
                               'constant', constant_values=0)
            elif image.ndim == 2:
                image = np.pad(image, ((padding[0], padding[0]), (padding[1], padding[1])),
                               'constant', constant_values=0)
            else:
                raise ValueError("Image must be 2D or 3D.")
        else:
            if image.ndim == 4:
                image = np.pad(image, ((0, 0), (padding[0], padding[0]), (padding[1], padding[1]), (0, 0)),
                               'constant', constant_values=0)
            elif image.ndim == 3:
                image = np.pad(image, ((0, 0), (padding[0], padding[0]), (padding[1], padding[1])),
                               'constant', constant_values=0)
            else:
                raise ValueError("Image must be 2D or 3D.")

        return image

    def __call__(self, image: np.ndarray, kernel: np.ndarray) -> np.ndarray:
        return self.convolve(image, kernel)

    def __str__(self):
        return f"{self.__class__.__name__}({self.kernel_size}, {self.padding}, {self.stride})"



from .activation_function import ActivationFunction
from .sigmoid import Sigmoid
from .softmax import Softmax
from .relu import ReLU
from .linear import Linear
from .tanh import Tanh
from .utils import get_activation_function
